use crate::{
    apub::actions::{AnnounceComment, CreateComment, DeleteComment, DeleteReact, UpdateComment},
    Action, Context, Error, Required,
};

impl Action for CreateComment {
    fn perform(&self, context: &Context) -> Result<(), Error> {
        let comment = context.store.comments.create(
            self.submission_id,
            self.profile_id,
            self.comment_id,
            &self.body,
            self.published,
        )?;

        context.apub.comment(&self.apub_id, comment.id())?;

        if let Some(comment_id) = self.comment_id {
            let commenter_id = context
                .store
                .comments
                .by_id(comment_id)?
                .req()?
                .profile_id();

            if let Ok(Some(true)) = context.store.profiles.is_local(commenter_id) {
                context
                    .store
                    .view
                    .comments
                    .new(commenter_id, comment.id(), self.published);
            }
        };

        let submissioner_id = context
            .store
            .submissions
            .by_id(self.submission_id)?
            .req()?
            .profile_id();

        if let Ok(Some(true)) = context.store.profiles.is_local(submissioner_id) {
            context
                .store
                .view
                .comments
                .new(submissioner_id, comment.id(), self.published);
        }

        Ok(())
    }
}

impl Action for AnnounceComment {
    fn perform(&self, _: &Context) -> Result<(), Error> {
        Ok(())
    }
}

impl Action for UpdateComment {
    fn perform(&self, context: &Context) -> Result<(), Error> {
        let comment = context.store.comments.by_id(self.comment_id)?.req()?;

        let mut changes = comment.update();
        if let Some(body) = &self.body {
            changes.body(body);
        }
        context.store.comments.update(&changes)?;

        Ok(())
    }
}

impl Action for DeleteComment {
    fn perform(&self, context: &Context) -> Result<(), Error> {
        let opt = context.store.comments.delete(self.comment_id)?;
        context.store.view.comments.remove(self.comment_id);

        if opt.is_some() {
            let comment_id = self.comment_id;
            let context_clone = context.clone();

            context.arbiter.send(Box::pin(async move {
                let _ = actix_web::web::block(move || {
                    for react_id in context_clone.store.reacts.for_comment(comment_id) {
                        if let Err(e) = Action::perform(&DeleteReact { react_id }, &context_clone) {
                            log::error!("Error deleing react {}: {}", react_id, e);
                        }
                    }
                    Ok(()) as Result<(), ()>
                })
                .await;
            }));
        }

        Ok(())
    }
}
