use actix_web::{body::BodyStream, client::Client, dev::Payload, HttpRequest, HttpResponse};
use std::{fmt, rc::Rc};
use url::Url;

#[derive(Debug, Clone)]
struct Serde<T> {
    inner: T,
}

impl<T> Serde<T> {
    pub(crate) fn new(inner: T) -> Self {
        Serde { inner }
    }
}

impl<T> serde::Serialize for Serde<T>
where
    T: std::fmt::Display,
{
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let s = self.inner.to_string();
        serde::Serialize::serialize(s.as_str(), serializer)
    }
}

impl<'de, T> serde::Deserialize<'de> for Serde<T>
where
    T: std::str::FromStr,
    <T as std::str::FromStr>::Err: std::fmt::Display,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let s: String = serde::Deserialize::deserialize(deserializer)?;
        let inner = s
            .parse::<T>()
            .map_err(|e| serde::de::Error::custom(e.to_string()))?;

        Ok(Serde { inner })
    }
}

#[derive(Debug, serde::Deserialize)]
struct Details {
    width: usize,
    height: usize,
    content_type: Serde<mime::Mime>,
}

#[derive(Debug, serde::Deserialize)]
pub struct Image {
    file: String,
    delete_token: String,
    details: Details,
}

impl Image {
    pub(crate) fn key(&self) -> &str {
        &self.file
    }

    pub(crate) fn token(&self) -> &str {
        &self.delete_token
    }
}

#[derive(Debug, serde::Deserialize)]
pub struct Images {
    msg: String,
    files: Option<Vec<Image>>,
}

impl Images {
    pub(crate) fn is_err(&self) -> bool {
        self.files.is_none()
    }

    pub(crate) fn message(&self) -> &str {
        &self.msg
    }

    pub(crate) fn images(&self) -> impl DoubleEndedIterator<Item = &Image> {
        self.files.iter().flat_map(|v| v.iter())
    }
}

#[derive(Debug, thiserror::Error)]
pub enum UploadError {
    #[error("Failed to send request")]
    Request,

    #[error("Failed to parse image json")]
    Json,
}

pub trait ImageInfo {
    fn icon_sizes(&self) -> &[u64];
    fn banner_sizes(&self) -> &[u64];
    fn thumbnail_sizes(&self) -> &[u64];

    fn icon_url(&self, key: &str, kind: ImageType, size: u64) -> Url;
    fn banner_url(&self, key: &str, kind: ImageType, size: u64) -> Url;
    fn thumbnail_url(&self, key: &str, kind: ImageType, size: u64) -> Url;
    fn image_url(&self, key: &str) -> Url;
}

#[derive(Clone, Copy, Debug, serde::Deserialize, serde::Serialize)]
pub enum ImageType {
    #[serde(rename = "jpg")]
    Jpeg,

    #[serde(rename = "png")]
    Png,

    #[serde(rename = "webp")]
    Webp,
}

#[derive(Clone)]
pub struct State {
    upstream: Url,
    image_info: Rc<dyn ImageInfo>,
    client: Client,
}

impl State {
    pub(crate) fn new(upstream: Url, image_info: impl ImageInfo + 'static, client: Client) -> Self {
        State {
            upstream,
            image_info: Rc::new(image_info),
            client,
        }
    }

    pub(super) fn image_info(&self) -> &dyn ImageInfo {
        &*self.image_info
    }

    pub(crate) async fn serve_thumbnail(
        &self,
        req: &HttpRequest,
        key: &str,
        kind: ImageType,
        size: u64,
    ) -> Result<HttpResponse, actix_web::Error> {
        self.serve_image(self.thumbnail(key, kind, size), req).await
    }

    pub(crate) async fn serve_banner(
        &self,
        req: &HttpRequest,
        key: &str,
        kind: ImageType,
        size: u64,
    ) -> Result<HttpResponse, actix_web::Error> {
        self.serve_image(self.banner(key, kind, size), req).await
    }

    pub(crate) async fn serve_icon(
        &self,
        req: &HttpRequest,
        key: &str,
        kind: ImageType,
        size: u64,
    ) -> Result<HttpResponse, actix_web::Error> {
        self.serve_image(self.icon(key, kind, size), req).await
    }

    pub(crate) async fn serve_full(
        &self,
        req: &HttpRequest,
        key: &str,
    ) -> Result<HttpResponse, actix_web::Error> {
        self.serve_image(self.full_size(key), req).await
    }

    pub(crate) async fn proxy_upload(
        &self,
        req: HttpRequest,
        body: Payload,
    ) -> Result<Images, UploadError> {
        let client_req = self.client.request_from(self.upload().as_str(), req.head());

        let client_req = if let Some(addr) = req.head().peer_addr {
            client_req.header("X-Forwarded-For", addr.to_string())
        } else {
            client_req
        };

        let mut res = client_req
            .send_stream(body)
            .await
            .map_err(|_| UploadError::Request)?;

        let images = res.json::<Images>().await.map_err(|_| UploadError::Json)?;

        Ok(images)
    }

    async fn serve_image(
        &self,
        url: Url,
        req: &HttpRequest,
    ) -> Result<HttpResponse, actix_web::Error> {
        let client_req = self.client.request_from(url.as_str(), req.head());
        let client_req = if let Some(addr) = req.head().peer_addr {
            client_req.header("X-Forwarded-For", addr.to_string())
        } else {
            client_req
        };

        let res = client_req.no_decompress().send().await?;

        let mut client_res = HttpResponse::build(res.status());

        for (name, value) in res.headers().iter().filter(|(h, _)| *h != "connection") {
            client_res.header(name.clone(), value.clone());
        }

        Ok(client_res.body(BodyStream::new(res)))
    }

    pub(crate) fn upload(&self) -> Url {
        let mut url = self.upstream.clone();
        url.set_path("/image");
        url
    }

    pub(crate) fn full_size(&self, key: &str) -> Url {
        let mut url = self.upstream.clone();
        url.set_path(&format!("/image/original/{}", key));
        url
    }

    fn process(&self, kind: ImageType) -> Url {
        let mut url = self.upstream.clone();
        url.set_path(&format!("/image/process.{}", kind));
        url
    }

    pub(crate) fn icon(&self, key: &str, kind: ImageType, size: u64) -> Url {
        let mut url = self.process(kind);
        url.set_query(Some(&format!("src={}&crop=1x1&resize={}", key, size)));
        url
    }

    pub(crate) fn banner(&self, key: &str, kind: ImageType, size: u64) -> Url {
        let mut url = self.process(kind);
        url.set_query(Some(&format!("src={}&crop=3x1&resize={}", key, size)));
        url
    }

    pub(crate) fn thumbnail(&self, key: &str, kind: ImageType, size: u64) -> Url {
        let mut url = self.process(kind);
        url.set_query(Some(&format!("src={}&resize={}", key, size)));
        url
    }

    pub(crate) fn delete(&self, key: &str, token: &str) -> Url {
        let mut url = self.upstream.clone();
        url.set_path(&format!("/image/delete/{}/{}", token, key));
        url
    }
}

impl fmt::Display for ImageType {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let s = match self {
            Self::Png => "png",
            Self::Jpeg => "jpg",
            Self::Webp => "webp",
        };

        write!(f, "{}", s)
    }
}

impl fmt::Debug for State {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_struct("State")
            .field("upstream", &self.upstream)
            .field("image_info", &"Rc<dyn ImageInfo>")
            .field("client", &"Client")
            .finish()
    }
}
