use super::{Owner, OwnerSource, StoreError, Undo};
use chrono::{DateTime, Utc};
use sled::{Db, Transactional, Tree};
use std::io::Cursor;
use uuid::Uuid;

#[derive(Clone, Debug)]
pub struct Store {
    owner_tree: Tree,
    source_tree: Tree,
}

#[derive(Clone, Debug, serde::Deserialize, serde::Serialize)]
enum StoredOwnerSource<'a> {
    Local(Uuid),
    Remote(&'a str),
}

#[derive(Debug, serde::Deserialize, serde::Serialize)]
struct StoredOwner<'a> {
    id: Uuid,
    #[serde(borrow)]
    source: StoredOwnerSource<'a>,
    created_at: DateTime<Utc>,
    updated_at: DateTime<Utc>,
}

impl Store {
    pub fn build(db: &Db) -> Result<Self, sled::Error> {
        Ok(Store {
            owner_tree: db.open_tree("profiles/owners")?,
            source_tree: db.open_tree("profiles/owners/source")?,
        })
    }

    pub fn create(&self, source: &OwnerSource) -> Result<Owner, StoreError> {
        let mut id;
        let mut stored_owner;

        let now = Utc::now().into();
        let mut stored_owner_vec = vec![];

        let stored_source = match source {
            OwnerSource::Local(id) => StoredOwnerSource::Local(*id),
            OwnerSource::Remote(s) => StoredOwnerSource::Remote(&s),
        };

        while {
            stored_owner_vec.clear();
            let writer = Cursor::new(&mut stored_owner_vec);
            id = Uuid::new_v4();
            stored_owner = StoredOwner {
                id,
                source: stored_source.clone(),
                created_at: now,
                updated_at: now,
            };
            serde_json::to_writer(writer, &stored_owner)?;
            self.owner_tree
                .compare_and_swap(
                    id_owner_key(id),
                    None as Option<&[u8]>,
                    Some(stored_owner_vec.as_slice()),
                )?
                .is_err()
        } {}

        let source_key = source_key(stored_source.clone());

        if let Err(e) = self
            .source_tree
            .insert(source_key, id.to_string().as_bytes())
        {
            self.owner_tree.remove(id_owner_key(id))?;
            return Err(e.into());
        }

        Ok(stored_owner.into())
    }

    pub fn by_source(&self, source: &OwnerSource) -> Result<Option<Owner>, StoreError> {
        let stored_source = match source {
            OwnerSource::Local(id) => StoredOwnerSource::Local(*id),
            OwnerSource::Remote(s) => StoredOwnerSource::Remote(&s),
        };

        let source_key = source_key(stored_source);

        let id_ivec = match self.source_tree.get(source_key)? {
            Some(ivec) => ivec,
            None => return Ok(None),
        };

        let id_str = String::from_utf8_lossy(&id_ivec);
        let id: Uuid = id_str.parse().expect("Uuid is valid");

        let owner_ivec = match self.owner_tree.get(id_owner_key(id))? {
            Some(ivec) => ivec,
            None => return Ok(None),
        };

        let stored_owner: StoredOwner = serde_json::from_slice(&owner_ivec)?;

        Ok(Some(stored_owner.into()))
    }

    pub fn delete(&self, owner_id: Uuid) -> Result<Option<Undo<Owner>>, StoreError> {
        let owner_ivec = match self.owner_tree.get(id_owner_key(owner_id))? {
            Some(ivec) => ivec,
            None => return Ok(None),
        };

        let stored_owner: StoredOwner = serde_json::from_slice(&owner_ivec)?;

        let id = owner_id;
        let source_key = source_key(stored_owner.source.clone());

        let owner = stored_owner.into();

        &[&self.owner_tree, &self.source_tree].transaction(move |trees| {
            let owner_tree = &trees[0];
            let source_tree = &trees[1];

            source_tree.remove(source_key.as_bytes())?;
            owner_tree.remove(id_owner_key(id).as_bytes())?;

            Ok(())
        })?;

        Ok(Some(Undo(owner)))
    }
}

fn id_owner_key(id: Uuid) -> String {
    format!("/owner/{}/data", id)
}

fn source_key<'a>(source: StoredOwnerSource<'a>) -> String {
    match source {
        StoredOwnerSource::Local(id) => format!("/source/local:{}", id),
        StoredOwnerSource::Remote(s) => format!("/source/remote:{}", s),
    }
}

impl<'a> From<StoredOwner<'a>> for Owner {
    fn from(so: StoredOwner<'a>) -> Self {
        Owner {
            id: so.id,
            source: match so.source {
                StoredOwnerSource::Local(id) => OwnerSource::Local(id),
                StoredOwnerSource::Remote(s) => OwnerSource::Remote(s.to_owned()),
            },
        }
    }
}
