use std::{cell::RefCell, fmt};

#[derive(Clone, Copy, Debug)]
pub enum ButtonKind {
    Primary,
    PrimaryOutline,
    Outline,
    Secondary,
}

#[derive(Clone, Copy, Debug)]
pub enum LinkKind {
    CurrentTab,
    NewTab,
}

#[derive(Default)]
pub struct Button {
    pub(crate) inner: RefCell<Inner>,
}

impl fmt::Debug for Button {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_struct("Button").field("inner", &"Inner").finish()
    }
}

#[derive(Default)]
pub(crate) struct Inner {
    pub(crate) label: String,
    pub(crate) kind: ButtonKind,
    pub(crate) classes: Vec<String>,
    pub(crate) href: Option<String>,
    pub(crate) link_kind: LinkKind,
    pub(crate) action: Option<String>,
}

impl Button {
    pub fn primary(label: &str) -> Self {
        Button {
            inner: RefCell::new(Inner {
                label: label.to_owned(),
                ..Inner::default()
            }),
        }
    }

    pub fn primary_outline(label: &str) -> Self {
        Button {
            inner: RefCell::new(Inner {
                label: label.to_owned(),
                kind: ButtonKind::PrimaryOutline,
                ..Inner::default()
            }),
        }
    }

    pub fn outline(label: &str) -> Self {
        Button {
            inner: RefCell::new(Inner {
                label: label.to_owned(),
                kind: ButtonKind::Outline,
                ..Inner::default()
            }),
        }
    }

    pub fn secondary(label: &str) -> Self {
        Button {
            inner: RefCell::new(Inner {
                label: label.to_owned(),
                kind: ButtonKind::Secondary,
                ..Inner::default()
            }),
        }
    }

    pub fn kind(&self, kind: ButtonKind) -> &Self {
        self.inner.borrow_mut().kind = kind;
        self
    }

    pub fn classes(&self, classes: &[&str]) -> &Self {
        self.inner.borrow_mut().classes = classes.into_iter().map(|s| s.to_string()).collect();
        self
    }

    pub fn href(&self, href: &str) -> &Self {
        self.inner.borrow_mut().href = Some(href.to_owned());
        self
    }

    pub fn new_tab(&self) -> &Self {
        self.inner.borrow_mut().link_kind = LinkKind::NewTab;
        self
    }

    pub fn form(&self, action: &str) -> &Self {
        self.inner.borrow_mut().action = Some(action.to_owned());
        self
    }

    pub(crate) fn class_string(&self) -> String {
        use ButtonKind::*;

        let static_classes = match self.inner.borrow().kind {
            Primary => "toolkit-button toolkit-button__primary",
            PrimaryOutline => "toolkit-button toolkit-button__primary-outline",
            Outline => "toolkit-button toolkit-button__outline",
            Secondary => "toolkit-button toolkit-button__secondary",
        };

        self.inner.borrow().classes.join(" ") + " " + static_classes
    }
}

impl Default for ButtonKind {
    fn default() -> Self {
        ButtonKind::Primary
    }
}

impl Default for LinkKind {
    fn default() -> Self {
        LinkKind::CurrentTab
    }
}
